@@ -40,11 +40,7 @@
  * https://terminal-wg.pages.freedesktop.org/bidi/
  */
 
-#include <config.h>
-
-#if WITH_FRIBIDI
-#include <fribidi.h>
-#endif
+#include "config.h"
 
 #include "bidi.hh"
 #include "debug.h"
@@ -53,6 +49,7 @@
 
 #if WITH_FRIBIDI
 static_assert (sizeof (FriBidiChar) == sizeof (gunichar), "Unexpected FriBidiChar size");
+static_assert (sizeof (FriBidiStrIndex) == sizeof (int), "Unexpected FriBidiStrIndex size");
 #endif
 
 /* Don't do Arabic ligatures as per bug 142. */
@@ -71,7 +68,7 @@ BidiRow::~BidiRow()
 void
 BidiRow::set_width(vte::grid::column_t width)
 {
-        g_assert_cmpint(width, >=, 0);
+        vte_assert_cmpint(width, >=, 0);
         if (G_UNLIKELY (width > G_MAXUSHORT)) {
                 width = G_MAXUSHORT;
         }
@@ -169,7 +166,7 @@ BidiRow::log_is_rtl(vte::grid::column_t col) const
 vteunistr
 BidiRow::vis_get_shaped_char(vte::grid::column_t col, vteunistr s) const
 {
-        g_assert_cmpint (col, >=, 0);
+        vte_assert_cmpint (col, >=, 0);
 
         if (col >= m_width || m_vis_shaped_base_char[col] == 0)
                 return s;
@@ -205,15 +202,9 @@ BidiRunner::explicit_line_shape(vte::grid::row_t row)
 
         auto width = m_ringview->get_width();
 
-        GArray *fribidi_chars_array = nullptr;
-
         FriBidiParType pbase_dir = FRIBIDI_PAR_RTL;
         FriBidiLevel level;
         FriBidiChar *fribidi_chars;
-        std::vector<FriBidiCharType> fribidi_chartypes;
-        std::vector<FriBidiBracketType> fribidi_brackettypes;
-        std::vector<FriBidiJoiningType> fribidi_joiningtypes;
-        std::vector<FriBidiLevel> fribidi_levels;
 
         int count;
 
@@ -222,7 +213,11 @@ BidiRunner::explicit_line_shape(vte::grid::row_t row)
         gunichar base;
         int i, j;  /* visual columns */
 
-        fribidi_chars_array = g_array_sized_new (FALSE, FALSE, sizeof (FriBidiChar), width);
+        VteBidiChars *fribidi_chars_array = &m_fribidi_chars_array;
+        vte_bidi_chars_set_size(fribidi_chars_array, 0);
+        vte_bidi_chars_reserve(fribidi_chars_array, width);
+
+        static thread_local auto workspace = Workspace{};
 
         /* Walk in visual order from right to left. */
         i = width - 1;
@@ -236,12 +231,12 @@ BidiRunner::explicit_line_shape(vte::grid::row_t row)
                 }
 
                 /* Found an Arabic character. Keep walking to the left, extracting the word. */
-                g_array_set_size(fribidi_chars_array, 0);
+                vte_bidi_chars_set_size(fribidi_chars_array, 0);
                 j = i;
                 do {
-                        auto prev_len = fribidi_chars_array->len;
+                        G_GNUC_UNUSED auto prev_len = vte_bidi_chars_get_size (fribidi_chars_array);
                         _vte_unistr_append_to_gunichars (cell->c, fribidi_chars_array);
-                        g_assert_cmpint (fribidi_chars_array->len, >, prev_len);
+                        vte_assert_cmpint (vte_bidi_chars_get_size (fribidi_chars_array), >, prev_len);
 
                         j--;
                         if (j >= 0) {
@@ -256,20 +251,22 @@ BidiRunner::explicit_line_shape(vte::grid::row_t row)
 
                 /* Extracted the Arabic run. Do the BiDi. */
 
-                /* Convenience stuff, we no longer need the auto-growing GArray wrapper. */
-                count = fribidi_chars_array->len;
-                fribidi_chars = reinterpret_cast<FriBidiChar*>(fribidi_chars_array->data);
+                /* Convenience stuff, we no longer need the auto-growing wrapper. */
+                count = vte_bidi_chars_get_size (fribidi_chars_array);
+                fribidi_chars = vte_bidi_chars_get_data (fribidi_chars_array);
+
+                workspace.reserve(count);
 
                 /* Run the BiDi algorithm on the paragraph to get the embedding levels. */
-                fribidi_chartypes.resize(count);
-                fribidi_brackettypes.resize(count);
-                fribidi_joiningtypes.resize(count);
-                fribidi_levels.resize(count);
+                auto fribidi_chartypes = workspace.char_types_data();
+                auto fribidi_brackettypes = workspace.bracket_types_data();
+                auto fribidi_joiningtypes = workspace.joining_types_data();
+                auto fribidi_levels = workspace.levels_data();
 
-                fribidi_get_bidi_types (fribidi_chars, count, fribidi_chartypes.data());
-                fribidi_get_bracket_types (fribidi_chars, count, fribidi_chartypes.data(), fribidi_brackettypes.data());
-                fribidi_get_joining_types (fribidi_chars, count, fribidi_joiningtypes.data());
-                level = fribidi_get_par_embedding_levels_ex (fribidi_chartypes.data(), fribidi_brackettypes.data(), count, &pbase_dir, fribidi_levels.data()) - 1;
+                fribidi_get_bidi_types (fribidi_chars, count, fribidi_chartypes);
+                fribidi_get_bracket_types (fribidi_chars, count, fribidi_chartypes, fribidi_brackettypes);
+                fribidi_get_joining_types (fribidi_chars, count, fribidi_joiningtypes);
+                level = fribidi_get_par_embedding_levels_ex (fribidi_chartypes, fribidi_brackettypes, count, &pbase_dir, fribidi_levels) - 1;
                 if (level == (FriBidiLevel)(-1)) {
                         /* Error. Skip shaping this word. */
                         i = j - 1;
@@ -277,8 +274,8 @@ BidiRunner::explicit_line_shape(vte::grid::row_t row)
                 }
 
                 /* Shaping. */
-                fribidi_join_arabic (fribidi_chartypes.data(), count, fribidi_levels.data(), fribidi_joiningtypes.data());
-                fribidi_shape_arabic (VTE_ARABIC_SHAPING_FLAGS, fribidi_levels.data(), count, fribidi_joiningtypes.data(), fribidi_chars);
+                fribidi_join_arabic (fribidi_chartypes, count, fribidi_levels, fribidi_joiningtypes);
+                fribidi_shape_arabic (VTE_ARABIC_SHAPING_FLAGS, fribidi_levels, count, fribidi_joiningtypes, fribidi_chars);
 
                 /* If we have the shortcut notation for the trivial LTR mapping, we need to
                  * expand it to the nontrivial notation, in order to store the shaped character. */
@@ -294,7 +291,7 @@ BidiRunner::explicit_line_shape(vte::grid::row_t row)
                 /* Walk through the Arabic word again. */
                 j = i;
                 while (count > 0) {
-                        g_assert_cmpint (j, >=, 0);
+                        vte_assert_cmpint (j, >=, 0);
                         cell = _vte_row_data_get(row_data, bidirow->vis2log(j));
                         c = cell->c;
                         base = _vte_unistr_get_base(c);
@@ -312,7 +309,7 @@ BidiRunner::explicit_line_shape(vte::grid::row_t row)
                 i = j - 1;
         }
 
-        g_array_free (fribidi_chars_array, TRUE);
+        vte_bidi_chars_set_size (fribidi_chars_array, 0);
 }
 #endif /* WITH_FRIBIDI */
 
@@ -414,13 +411,16 @@ BidiRunner::implicit_paragraph(vte::grid::row_t start, vte::grid::row_t end, boo
         FriBidiParType pbase_dir;
         FriBidiLevel level;
         FriBidiChar *fribidi_chars;
-        std::vector<FriBidiCharType> fribidi_chartypes;
-        std::vector<FriBidiBracketType> fribidi_brackettypes;
-        std::vector<FriBidiJoiningType> fribidi_joiningtypes;
-        std::vector<FriBidiLevel> fribidi_levels;
+        FriBidiCharType *fribidi_chartypes;
+        FriBidiBracketType *fribidi_brackettypes;
+        FriBidiJoiningType *fribidi_joiningtypes;
+        FriBidiLevel *fribidi_levels;
         FriBidiStrIndex *fribidi_map;
         FriBidiStrIndex *fribidi_to_term;
         BidiRow *bidirow;
+        VteBidiChars *fribidi_chars_array;
+        VteBidiIndexes *fribidi_map_array;
+        VteBidiIndexes *fribidi_to_term_array;
 
         auto width = m_ringview->get_width();
 
@@ -436,9 +436,17 @@ BidiRunner::implicit_paragraph(vte::grid::row_t start, vte::grid::row_t end, boo
         int fl, fv;     /* fribidi logical and visual */
         unsigned int col;
 
-        GArray *fribidi_chars_array   = g_array_sized_new (FALSE, FALSE, sizeof (FriBidiChar),     (end - start) * width);
-        GArray *fribidi_map_array     = g_array_sized_new (FALSE, FALSE, sizeof (FriBidiStrIndex), (end - start) * width);
-        GArray *fribidi_to_term_array = g_array_sized_new (FALSE, FALSE, sizeof (FriBidiStrIndex), (end - start) * width);
+        fribidi_chars_array = &m_fribidi_chars_array;
+        vte_bidi_chars_set_size (fribidi_chars_array, 0);
+        vte_bidi_chars_reserve (fribidi_chars_array, (end - start) * width);
+
+        fribidi_map_array = &m_fribidi_map_array;
+        vte_bidi_indexes_set_size (fribidi_map_array, 0);
+        vte_bidi_indexes_reserve (fribidi_map_array, (end - start) * width);
+
+        fribidi_to_term_array = &m_fribidi_to_term_array;
+        vte_bidi_indexes_set_size (fribidi_to_term_array, 0);
+        vte_bidi_indexes_reserve (fribidi_to_term_array, (end - start) * width);
 
         /* Extract the paragraph's contents, omitting unused and fragment cells. */
 
@@ -525,7 +533,7 @@ BidiRunner::implicit_paragraph(vte::grid::row_t start, vte::grid::row_t end, boo
                 row_data = m_ringview->get_row(row);
 
                 for (tl = 0; tl < row_data->len; tl++) {
-                        auto prev_len = fribidi_chars_array->len;
+                        auto prev_len = vte_bidi_chars_get_size(fribidi_chars_array);
                         FriBidiStrIndex val;
 
                         cell = _vte_row_data_get (row_data, tl);
@@ -537,68 +545,71 @@ BidiRunner::implicit_paragraph(vte::grid::row_t start, vte::grid::row_t end, boo
                          * Note: see the static assert at the top of this file. */
                         _vte_unistr_append_to_gunichars (cell->c ? cell->c : ' ', fribidi_chars_array);
                         /* Make sure at least one character was produced. */
-                        g_assert_cmpint (fribidi_chars_array->len, >, prev_len);
+                        vte_assert_cmpint (vte_bidi_chars_get_size(fribidi_chars_array), >, prev_len);
 
                         /* Track the base character, assign to it its current index in fribidi_chars.
                          * Don't track combining accents, assign -1's to them. */
                         val = prev_len;
-                        g_array_append_val (fribidi_map_array, val);
+                        vte_bidi_indexes_append (fribidi_map_array, &val);
                         val = tl;
-                        g_array_append_val (fribidi_to_term_array, val);
+                        vte_bidi_indexes_append (fribidi_to_term_array, &val);
                         prev_len++;
                         val = -1;
-                        while (prev_len++ < fribidi_chars_array->len) {
-                                g_array_append_val (fribidi_map_array, val);
-                                g_array_append_val (fribidi_to_term_array, val);
+                        while (prev_len++ < vte_bidi_chars_get_size(fribidi_chars_array)) {
+                                vte_bidi_indexes_append (fribidi_map_array, &val);
+                                vte_bidi_indexes_append (fribidi_to_term_array, &val);
                         }
                 }
 
-                lines[++line] = fribidi_chars_array->len;
+                lines[++line] = vte_bidi_chars_get_size(fribidi_chars_array);
         }
 
-        /* Convenience stuff, we no longer need the auto-growing GArray wrapper. */
-        count = fribidi_chars_array->len;
-        fribidi_chars = reinterpret_cast<FriBidiChar*>(fribidi_chars_array->data);
-        fribidi_map = reinterpret_cast<FriBidiStrIndex*>(fribidi_map_array->data);
-        fribidi_to_term = reinterpret_cast<FriBidiStrIndex*>(fribidi_to_term_array->data);
+        /* Convenience stuff, we no longer need the auto-growing wrapper. */
+        count = vte_bidi_chars_get_size(fribidi_chars_array);
+        fribidi_chars = vte_bidi_chars_get_data(fribidi_chars_array);
+        fribidi_map = vte_bidi_indexes_get_data(fribidi_map_array);
+        fribidi_to_term = vte_bidi_indexes_get_data(fribidi_to_term_array);
 
         /* Run the BiDi algorithm on the paragraph to get the embedding levels. */
-        fribidi_chartypes.resize(count);
-        fribidi_brackettypes.resize(count);
-        fribidi_joiningtypes.resize(count);
-        fribidi_levels.resize(count);
+        static thread_local auto workspace = Workspace{};
+        workspace.reserve(count);
+
+        fribidi_chartypes = workspace.char_types_data();
+        fribidi_brackettypes = workspace.bracket_types_data();
+        fribidi_joiningtypes = workspace.joining_types_data();
+        fribidi_levels = workspace.levels_data();
 
         pbase_dir = autodir ? (rtl ? FRIBIDI_PAR_WRTL : FRIBIDI_PAR_WLTR)
                             : (rtl ? FRIBIDI_PAR_RTL  : FRIBIDI_PAR_LTR );
 
-        fribidi_get_bidi_types (fribidi_chars, count, fribidi_chartypes.data());
-        fribidi_get_bracket_types (fribidi_chars, count, fribidi_chartypes.data(), fribidi_brackettypes.data());
-        fribidi_get_joining_types (fribidi_chars, count, fribidi_joiningtypes.data());
-        level = fribidi_get_par_embedding_levels_ex (fribidi_chartypes.data(), fribidi_brackettypes.data(), count, &pbase_dir, fribidi_levels.data()) - 1;
+        fribidi_get_bidi_types (fribidi_chars, count, fribidi_chartypes);
+        fribidi_get_bracket_types (fribidi_chars, count, fribidi_chartypes, fribidi_brackettypes);
+        fribidi_get_joining_types (fribidi_chars, count, fribidi_joiningtypes);
+        level = fribidi_get_par_embedding_levels_ex (fribidi_chartypes, fribidi_brackettypes, count, &pbase_dir, fribidi_levels) - 1;
 
         if (level == (FriBidiLevel)(-1)) {
                 /* error */
-                g_array_free (fribidi_chars_array, TRUE);
-                g_array_free (fribidi_map_array, TRUE);
-                g_array_free (fribidi_to_term_array, TRUE);
+                vte_bidi_chars_set_size (fribidi_chars_array, 0);
+                vte_bidi_indexes_set_size (fribidi_map_array, 0);
+                vte_bidi_indexes_set_size (fribidi_to_term_array, 0);
                 return false;
         }
 
         if (do_shaping) {
                 /* Arabic shaping (on the entire paragraph in a single run). */
-                fribidi_join_arabic (fribidi_chartypes.data(), count, fribidi_levels.data(), fribidi_joiningtypes.data());
-                fribidi_shape_arabic (VTE_ARABIC_SHAPING_FLAGS, fribidi_levels.data(), count, fribidi_joiningtypes.data(), fribidi_chars);
+                fribidi_join_arabic (fribidi_chartypes, count, fribidi_levels, fribidi_joiningtypes);
+                fribidi_shape_arabic (VTE_ARABIC_SHAPING_FLAGS, fribidi_levels, count, fribidi_joiningtypes, fribidi_chars);
         }
 
         /* For convenience, from now on this variable contains the resolved (i.e. possibly autodetected) value. */
-        g_assert_cmpint (pbase_dir, !=, FRIBIDI_PAR_ON);
+        vte_assert_cmpint (pbase_dir, !=, FRIBIDI_PAR_ON);
         rtl = (pbase_dir == FRIBIDI_PAR_RTL || pbase_dir == FRIBIDI_PAR_WRTL);
 
         if (!rtl && level == 0) {
                 /* Fast and memory saving shortcut for LTR-only paragraphs. */
-                g_array_free (fribidi_chars_array, TRUE);
-                g_array_free (fribidi_map_array, TRUE);
-                g_array_free (fribidi_to_term_array, TRUE);
+                vte_bidi_chars_set_size (fribidi_chars_array, 0);
+                vte_bidi_indexes_set_size (fribidi_map_array, 0);
+                vte_bidi_indexes_set_size (fribidi_to_term_array, 0);
                 explicit_paragraph (start, end, false, false);
                 return true;
         }
@@ -640,11 +651,11 @@ BidiRunner::implicit_paragraph(vte::grid::row_t start, vte::grid::row_t end, boo
                 row_data = m_ringview->get_row(row);
 
                 level = fribidi_reorder_line (FRIBIDI_FLAGS_DEFAULT,
-                                              fribidi_chartypes.data(),
+                                              fribidi_chartypes,
                                               lines[line + 1] - lines[line],
                                               lines[line],
                                               pbase_dir,
-                                              fribidi_levels.data(),
+                                              fribidi_levels,
                                               NULL,
                                               fribidi_map) - 1;
 
@@ -704,14 +715,14 @@ BidiRunner::implicit_paragraph(vte::grid::row_t start, vte::grid::row_t end, boo
                 }
                 if (!rtl) {
                         /* Unused cells on the right for LTR paragraphs */
-                        g_assert_cmpint (tv, ==, row_data->len);
+                        vte_assert_cmpint (tv, ==, row_data->len);
                         for (; tv < width; tv++) {
                                 bidirow->m_vis2log[tv] = tv;
                                 bidirow->m_vis_rtl[tv] = false;
                                 bidirow->m_vis_shaped_base_char[tv] = 0;
                         }
                 }
-                g_assert_cmpint (tv, ==, width);
+                vte_assert_cmpint (tv, ==, width);
 
                 /* From vis2log create the log2vis mapping too.
                  * In debug mode assert that we have a bijective mapping. */
@@ -727,14 +738,14 @@ BidiRunner::implicit_paragraph(vte::grid::row_t start, vte::grid::row_t end, boo
 
                 if (_vte_debug_on (VTE_DEBUG_BIDI)) {
                         for (tl = 0; tl < width; tl++) {
-                                g_assert_cmpint (bidirow->m_log2vis[tl], !=, -1);
+                                vte_assert_cmpint (bidirow->m_log2vis[tl], !=, -1);
                         }
                 }
         }
 
-        g_array_free (fribidi_chars_array, TRUE);
-        g_array_free (fribidi_map_array, TRUE);
-        g_array_free (fribidi_to_term_array, TRUE);
+        vte_bidi_chars_set_size (fribidi_chars_array, 0);
+        vte_bidi_indexes_set_size (fribidi_map_array, 0);
+        vte_bidi_indexes_set_size (fribidi_to_term_array, 0);
         return true;
 }
 #endif /* WITH_FRIBIDI */
@@ -781,3 +792,4 @@ vte_bidi_get_mirror_char (vteunistr unistr, gboolean mirror_box_drawing, vteunis
                 *out = unistr_mirrored;
         return unistr_mirrored == unistr;
 }
+
