@@ -395,7 +395,7 @@ sequences = [
                comment='double width double height line: bottom half'),
     seq_ESCAPE('DECSWL', '5', intermediates=(Intermediate.HASH,), flags=Flags.NOP,
                comment='single width single height line'),
-    seq_ESCAPE('DECBI', '6', flags=Flags.NOP,
+    seq_ESCAPE('DECBI', '6',
                comment='back index'),
     seq_ESCAPE('DECDWL', '6', intermediates=(Intermediate.HASH,), flags=Flags.NOP,
                comment='double width single height line'),
@@ -405,7 +405,7 @@ sequences = [
                comment='restore cursor'),
     seq_ESCAPE('DECALN', '8', intermediates=(Intermediate.HASH,),
                comment='screen alignment pattern'),
-    seq_ESCAPE('DECFI', '9', flags=Flags.NOP,
+    seq_ESCAPE('DECFI', '9',
                comment='forward index'),
     seq_ESCAPE('WYDHL_TH', ':', intermediates=(Intermediate.HASH,), flags=Flags.NOP,
                comment='single width double height line: top half'),
@@ -489,11 +489,11 @@ sequences = [
     # CSI sequences
     seq_CSI('ICH', '@',
             comment='insert character'),
-    seq_CSI('SL', '@', intermediates=(Intermediate.SPACE,), flags=Flags.NOP,
+    seq_CSI('SL', '@', intermediates=(Intermediate.SPACE,),
             comment='scroll left'),
     seq_CSI('CUU', 'A',
             comment='cursor up'),
-    seq_CSI('SR', 'A', intermediates=(Intermediate.SPACE,), flags=Flags.NOP,
+    seq_CSI('SR', 'A', intermediates=(Intermediate.SPACE,),
             comment='scroll right'),
     seq_CSI('CUD', 'B',
             comment='cursor down'),
@@ -756,8 +756,6 @@ sequences = [
             comment='select terminal id'),
     seq_CSI('DECCRTST', 'q', intermediates=(Intermediate.MINUS,), flags=Flags.NOP,
             comment='CRT saver time'),
-    seq_CSI('XTERM_VERSION', 'q', pintro=(ParameterIntro.GT,),
-            comment='request xterm version report'),
     seq_CSI('DECSTBM', 'r',
             comment='set top and bottom margins'),
     seq_CSI('DECSKCV', 'r', intermediates=(Intermediate.SPACE,), flags=Flags.NOP,
@@ -892,7 +890,7 @@ sequences = [
             comment='xterm pop SGR stack'),
     seq_CSI('DECSASD', '}', intermediates=(Intermediate.CASH,), flags=Flags.NOP,
             comment='select active status display'),
-    seq_CSI('DECIC', '}', intermediates=(Intermediate.SQUOTE,), flags=Flags.NOP,
+    seq_CSI('DECIC', '}', intermediates=(Intermediate.SQUOTE,),
             comment='insert column'),
     seq_CSI('DECATC', '}', intermediates=(Intermediate.COMMA,), flags=Flags.NOP,
             comment='alternate text color'),
@@ -902,7 +900,7 @@ sequences = [
             comment='terminal mode emulation'),
     seq_CSI('DECSSDT', '~', intermediates=(Intermediate.CASH,), flags=Flags.NOP,
             comment='select status display line type'),
-    seq_CSI('DECDC', '~', intermediates=(Intermediate.SQUOTE,), flags=Flags.NOP,
+    seq_CSI('DECDC', '~', intermediates=(Intermediate.SQUOTE,),
             comment='delete column'),
     seq_CSI('DECPS', '~', intermediates=(Intermediate.COMMA,), flags=Flags.NOP,
             comment='play sound'),
