@@ -25,6 +25,20 @@
 
 #include <string.h>
 
+#if WITH_SIXEL
+
+#include "cxx-utils.hh"
+
+/* We should be able to hold a single fullscreen 4K image at most.
+ * 35MiB equals 3840 * 2160 * 4 plus a little extra. */
+#define IMAGE_FAST_MEMORY_USED_MAX (35 * 1024 * 1024)
+
+/* Hard limit on number of images to keep around. This limits the impact
+ * of potential issues related to algorithmic complexity. */
+#define IMAGE_FAST_COUNT_MAX 4096
+
+#endif /* WITH_SIXEL */
+
 /*
  * Copy the common attributes from VteCellAttr to VteStreamCellAttr or vice versa.
  */
@@ -49,11 +63,11 @@ Ring::validate() const
 			m_start, m_end - m_start, m_end,
 			m_max, m_end - m_writable);
 
-	g_assert_cmpuint(m_start, <=, m_writable);
-	g_assert_cmpuint(m_writable, <=, m_end);
+	vte_assert_cmpuint(m_start, <=, m_writable);
+	vte_assert_cmpuint(m_writable, <=, m_end);
 
-	g_assert_cmpuint(m_end - m_start, <=, m_max);
-	g_assert_cmpuint(m_end - m_writable, <=, m_mask);
+	vte_assert_cmpuint(m_end - m_start, <=, m_max);
+	vte_assert_cmpuint(m_end - m_writable, <=, m_mask);
 }
 #else
 #define validate(...) do { } while(0)
@@ -190,6 +204,115 @@ Ring::hyperlink_maybe_gc(row_t increment)
                 hyperlink_gc();
 }
 
+#if WITH_SIXEL
+
+void
+Ring::image_gc_region() noexcept
+{
+        cairo_region_t *region = cairo_region_create();
+
+        for (auto rit = m_image_map.rbegin();
+             rit != m_image_map.rend();
+             ) {
+                auto const& image = rit->second;
+                auto const rect = cairo_rectangle_int_t{image->get_left(),
+                                                        image->get_top(),
+                                                        image->get_width(),
+                                                        image->get_height()};
+
+                if (cairo_region_contains_rectangle(region, &rect) == CAIRO_REGION_OVERLAP_IN) {
+                        /* vte::image::Image has been completely overdrawn; delete it */
+
+                        m_image_fast_memory_used -= image->resource_size();
+
+                        /* Apparently this is the cleanest way to erase() with a reverse iterator... */
+                        /* Unlink the image from m_image_by_top_map, then erase it from m_image_map */
+                        unlink_image_from_top_map(image.get());
+                        rit = image_map_type::reverse_iterator{m_image_map.erase(std::next(rit).base())};
+                        continue;
+                }
+
+                cairo_region_union_rectangle(region, &rect);
+                ++rit;
+        }
+
+        cairo_region_destroy(region);
+}
+
+void
+Ring::image_gc() noexcept
+{
+        while (m_image_fast_memory_used > IMAGE_FAST_MEMORY_USED_MAX ||
+               m_image_map.size() > IMAGE_FAST_COUNT_MAX) {
+                if (m_image_map.empty()) {
+                        /* If this happens, we've miscounted somehow. */
+                        break;
+                }
+
+                auto& image = m_image_map.begin()->second;
+                m_image_fast_memory_used -= image->resource_size();
+                unlink_image_from_top_map(image.get());
+                m_image_map.erase(m_image_map.begin());
+        }
+}
+
+void
+Ring::unlink_image_from_top_map(vte::image::Image const* image) noexcept
+{
+        auto [begin, end] = m_image_by_top_map.equal_range(image->get_top());
+
+        for (auto it = begin; it != end; ++it) {
+                if (it->second != image)
+                        continue;
+
+                m_image_by_top_map.erase(it);
+                break;
+        }
+}
+
+void
+Ring::rebuild_image_top_map() /* throws */
+{
+        m_image_by_top_map.clear();
+
+        for (auto it = m_image_map.begin(), end = m_image_map.end();
+             it != end;
+             ++it) {
+                auto const& image = it->second;
+                m_image_by_top_map.emplace(std::piecewise_construct,
+                                           std::forward_as_tuple(image->get_top()),
+                                           std::forward_as_tuple(image.get()));
+        }
+}
+
+bool
+Ring::rewrap_images_in_range(Ring::image_by_top_map_type::iterator& it,
+                             size_t text_start_ofs,
+                             size_t text_end_ofs,
+                             row_t new_row_index) noexcept
+{
+        for (auto const end = m_image_by_top_map.end();
+             it != end;
+             ++it) {
+                auto const& image = it->second;
+                auto ofs = CellTextOffset{};
+
+                if (!frozen_row_column_to_text_offset(image->get_top(), 0, &ofs))
+                        return false;
+
+                if (ofs.text_offset >= text_end_ofs)
+                        break;
+
+                if (ofs.text_offset >= text_start_ofs && ofs.text_offset < text_end_ofs) {
+                        image->set_top(new_row_index);
+                }
+        }
+
+        return true;
+}
+
+#endif /* WITH_SIXEL */
+
 /*
  * Find existing idx for the hyperlink or allocate a new one.
  *
@@ -239,7 +362,7 @@ Ring::get_hyperlink_idx_no_update_current(char const* hyperlink)
         }
 
         /* All allocated slots are in use. Gotta allocate a new one */
-        g_assert_cmpuint(m_hyperlink_highest_used_idx + 1, ==, m_hyperlinks->len);
+        vte_assert_cmpuint(m_hyperlink_highest_used_idx + 1, ==, m_hyperlinks->len);
 
         /* VTE_HYPERLINK_COUNT_MAX should be big enough for this not to happen under
            normal circumstances. Anyway, it's cheap to protect against extreme ones. */
@@ -257,7 +380,7 @@ Ring::get_hyperlink_idx_no_update_current(char const* hyperlink)
         str = g_string_new_len (hyperlink, len);
         g_ptr_array_add(m_hyperlinks, str);
 
-        g_assert_cmpuint(m_hyperlink_highest_used_idx + 1, ==, m_hyperlinks->len);
+        vte_assert_cmpuint(m_hyperlink_highest_used_idx + 1, ==, m_hyperlinks->len);
 
         return idx;
 }
@@ -303,7 +426,7 @@ Ring::freeze_row(row_t position,
 	record.attr_start_offset = _vte_stream_head(m_attr_stream);
 	record.is_ascii = 1;
 
-	g_string_set_size (buffer, 0);
+	g_string_truncate (buffer, 0);
 	for (i = 0, cell = row->cells; i < row->len; i++, cell++) {
 		VteCellAttr attr;
 		int num_chars;
@@ -448,7 +571,7 @@ Ring::thaw_row(row_t position,
 				if (!_vte_stream_read (m_attr_stream, record.attr_start_offset, (char *) &attr_change, sizeof (attr_change)))
 					return;
 				record.attr_start_offset += sizeof (attr_change);
-                                g_assert_cmpuint (attr_change.attr.hyperlink_length, <=, VTE_HYPERLINK_TOTAL_LENGTH_MAX);
+                                vte_assert_cmpuint (attr_change.attr.hyperlink_length, <=, VTE_HYPERLINK_TOTAL_LENGTH_MAX);
                                 if (attr_change.attr.hyperlink_length && !_vte_stream_read (m_attr_stream, record.attr_start_offset, hyperlink_readbuf, attr_change.attr.hyperlink_length))
                                         return;
                                 hyperlink_readbuf[attr_change.attr.hyperlink_length] = '\0';
@@ -528,7 +651,7 @@ Ring::thaw_row(row_t position,
 			/* Check the previous attr record. If its text ends where truncating, this attr record also needs to be removed. */
                         guint16 hyperlink_length;
                         if (_vte_stream_read (m_attr_stream, attr_stream_truncate_at - 2, (char *) &hyperlink_length, 2)) {
-                                g_assert_cmpuint (hyperlink_length, <=, VTE_HYPERLINK_TOTAL_LENGTH_MAX);
+                                vte_assert_cmpuint (hyperlink_length, <=, VTE_HYPERLINK_TOTAL_LENGTH_MAX);
                                 if (_vte_stream_read (m_attr_stream, attr_stream_truncate_at - 2 - hyperlink_length - sizeof (attr_change), (char *) &attr_change, sizeof (attr_change))) {
                                         if (records[0].text_start_offset == attr_change.text_end_offset) {
                                                 _vte_debug_print (VTE_DEBUG_RING, "... at attribute change\n");
@@ -546,7 +669,7 @@ Ring::thaw_row(row_t position,
                                         m_last_attr.hyperlink_idx = get_hyperlink_idx(hyperlink_readbuf);
                                 }
                                 if (_vte_stream_read (m_attr_stream, attr_stream_truncate_at - 2, (char *) &hyperlink_length, 2)) {
-                                        g_assert_cmpuint (hyperlink_length, <=, VTE_HYPERLINK_TOTAL_LENGTH_MAX);
+                                        vte_assert_cmpuint (hyperlink_length, <=, VTE_HYPERLINK_TOTAL_LENGTH_MAX);
                                         if (_vte_stream_read (m_attr_stream, attr_stream_truncate_at - 2 - hyperlink_length - sizeof (attr_change), (char *) &attr_change, sizeof (attr_change))) {
                                                 m_last_attr_text_start_offset = attr_change.text_end_offset;
                                         } else {
@@ -590,6 +713,13 @@ Ring::reset()
         m_start = m_writable = m_end;
         m_cached_row_num = (row_t)-1;
 
+#if WITH_SIXEL
+        m_image_by_top_map.clear();
+        m_image_map.clear();
+        m_next_image_priority = 0;
+        m_image_fast_memory_used = 0;
+#endif
+
         return m_end;
 }
 
@@ -688,13 +818,6 @@ Ring::get_hyperlink_at_position(row_t position,
         return idx;
 }
 
-VteRowData*
-Ring::index_writable(row_t position)
-{
-	ensure_writable(position);
-	return get_writable_index(position);
-}
-
 void
 Ring::freeze_one_row()
 {
@@ -714,7 +837,7 @@ Ring::thaw_one_row()
 {
 	VteRowData* row;
 
-	g_assert_cmpuint(m_start, <, m_writable);
+	vte_assert_cmpuint(m_start, <, m_writable);
 
 	ensure_writable_room();
 
@@ -799,19 +922,6 @@ Ring::ensure_writable_room()
 	g_free (old_array);
 }
 
-void
-Ring::ensure_writable(row_t position)
-{
-	if (G_LIKELY(position >= m_writable))
-		return;
-
-	_vte_debug_print(VTE_DEBUG_RING, "Ensure writable %lu.\n", position);
-
-        //FIXMEchpe surely this can be optimised
-	while (position < m_writable)
-		thaw_one_row();
-}
-
 /**
  * Ring::resize:
  * @max_rows: new maximum numbers of rows in the ring
@@ -883,8 +993,8 @@ Ring::insert(row_t position, guint8 bidi_flags)
 	ensure_writable(position);
 	ensure_writable_room();
 
-	g_assert_cmpuint (position, >=, m_writable);
-	g_assert_cmpuint (position, <=, m_end);
+	vte_assert_cmpuint (position, >=, m_writable);
+	vte_assert_cmpuint (position, <=, m_end);
 
         //FIXMEchpe WTF use better data structures!
 	tmp = *get_writable_index(m_end);
@@ -1011,7 +1121,7 @@ Ring::frozen_row_column_to_text_offset(row_t position,
 		/* go on */
 	}
 
-	g_assert_cmpuint(position, <, m_writable);
+	vte_assert_cmpuint(position, <, m_writable);
 	if (!read_row_record(&records[0], position))
 		return false;
 	if ((position + 1) * sizeof (records[0]) < _vte_stream_head(m_row_stream)) {
@@ -1020,7 +1130,16 @@ Ring::frozen_row_column_to_text_offset(row_t position,
 	} else
 		records[1].text_start_offset = _vte_stream_head(m_text_stream);
 
-	g_string_set_size (buffer, records[1].text_start_offset - records[0].text_start_offset);
+	offset->fragment_cells = 0;
+	offset->eol_cells = -1;
+	offset->text_offset = records[0].text_start_offset;
+
+        /* Save some work if we're in column 0. This holds true for images, whose column
+         * positions are disregarded for the purposes of wrapping. */
+        if (column == 0)
+                return true;
+
+        g_string_set_size (buffer, records[1].text_start_offset - records[0].text_start_offset);
 	if (!_vte_stream_read(m_text_stream, records[0].text_start_offset, buffer->str, buffer->len))
 		return false;
 
@@ -1032,8 +1151,6 @@ Ring::frozen_row_column_to_text_offset(row_t position,
 	/* row and buffer now contain the same text, in different representation */
 
 	/* count the number of characters up to the given column */
-	offset->fragment_cells = 0;
-	offset->eol_cells = -1;
 	num_chars = 0;
 	for (i = 0, cell = row->cells; i < row->len && i < column; i++, cell++) {
 		if (G_LIKELY (!cell->attr.fragment())) {
@@ -1054,7 +1171,7 @@ Ring::frozen_row_column_to_text_offset(row_t position,
 		off++;
 		if ((buffer->str[off] & 0xC0) != 0x80) num_chars--;
 	}
-	offset->text_offset = records[0].text_start_offset + off;
+	offset->text_offset += off;
 	return true;
 }
 
@@ -1086,7 +1203,7 @@ Ring::frozen_row_text_offset_to_column(row_t position,
 		return true;
 	}
 
-	g_assert_cmpuint(position, <, m_writable);
+	vte_assert_cmpuint(position, <, m_writable);
 	if (!read_row_record(&records[0], position))
 		return false;
 	if ((position + 1) * sizeof (records[0]) < _vte_stream_head(m_row_stream)) {
@@ -1106,8 +1223,8 @@ Ring::frozen_row_text_offset_to_column(row_t position,
          * if the ring ends in a soft wrapped line; see bug 181), the position we're about to
          * locate can be anywhere in the string, including just after its last character,
          * but not beyond that. */
-        g_assert_cmpuint(offset->text_offset, >=, records[0].text_start_offset);
-        g_assert_cmpuint(offset->text_offset, <=, records[0].text_start_offset + buffer->len);
+        vte_assert_cmpuint(offset->text_offset, >=, records[0].text_start_offset);
+        vte_assert_cmpuint(offset->text_offset, <=, records[0].text_start_offset + buffer->len);
 
 	row = index(position);
 
@@ -1166,6 +1283,9 @@ Ring::rewrap(column_t columns,
 	gsize paragraph_len;  /* excluding trailing '\n' */
 	gsize attr_offset;
 	gsize old_ring_end;
+#if WITH_SIXEL
+	auto image_it = m_image_by_top_map.begin();
+#endif
 
 	if (G_UNLIKELY(length() == 0))
 		return;
@@ -1301,6 +1421,14 @@ Ring::rewrap(column_t columns,
 							}
 						}
 
+#if WITH_SIXEL
+						if (!rewrap_images_in_range(image_it,
+                                                                            new_record.text_start_offset,
+                                                                            text_offset,
+                                                                            new_row_index))
+							goto err;
+#endif
+
 						new_row_index++;
 						new_record.text_start_offset = text_offset;
 						new_record.attr_start_offset = attr_offset;
@@ -1350,6 +1478,14 @@ Ring::rewrap(column_t columns,
 			}
 		}
 
+#if WITH_SIXEL
+		if (!rewrap_images_in_range(image_it,
+                                            new_record.text_start_offset,
+                                            paragraph_end_text_offset,
+                                            new_row_index))
+			goto err;
+#endif
+
 		new_row_index++;
 		paragraph_start_text_offset = paragraph_end_text_offset;
 	}
@@ -1370,8 +1506,11 @@ Ring::rewrap(column_t columns,
 		if (new_markers[i].row == -1)
 			new_markers[i].row = markers[i]->row - old_ring_end + m_end;
 		/* Convert byte offset into visual column */
-		if (!frozen_row_text_offset_to_column(new_markers[i].row, &marker_text_offsets[i], &new_markers[i].col))
-			goto err;
+                if (!frozen_row_text_offset_to_column(new_markers[i].row, &marker_text_offsets[i], &new_markers[i].col)) {
+                        /* This really shouldn't happen. It's too late to "goto err", the old stream is closed, the ring is updated.
+                         * It would be a bit cumbersome to refactor the code to still revert here. Choose a simple solution. */
+                        new_markers[i].col = 0;
+                }
 		_vte_debug_print(VTE_DEBUG_RING,
 				"Marker #%d new coords:  text_offset %" G_GSIZE_FORMAT "  fragment_cells %d  eol_cells %d  ->  row %ld  col %ld\n",
 				i, marker_text_offsets[i].text_offset, marker_text_offsets[i].fragment_cells,
@@ -1382,6 +1521,14 @@ Ring::rewrap(column_t columns,
 	g_free(marker_text_offsets);
 	g_free(new_markers);
 
+#if WITH_SIXEL
+        try {
+                rebuild_image_top_map();
+        } catch (...) {
+                vte::log_exception();
+        }
+#endif
+
 	_vte_debug_print(VTE_DEBUG_RING, "Ring after rewrapping:\n");
         validate();
 	return;
@@ -1412,7 +1559,7 @@ Ring::write_row(GOutputStream* stream,
 
 	/* Simple version of the loop in freeze_row().
 	 * TODO Should unify one day */
-	g_string_set_size (buffer, 0);
+	g_string_truncate (buffer, 0);
 	for (i = 0, cell = row->cells; i < row->len; i++, cell++) {
 		if (G_LIKELY (!cell->attr.fragment()))
 			_vte_unistr_append_to_string (cell->c, buffer);
@@ -1485,3 +1632,56 @@ Ring::write_contents(GOutputStream* stream,
 
 	return true;
 }
+
+#if WITH_SIXEL
+
+/**
+ * Ring::append_image:
+ * @surface: A Cairo surface object
+ * @pixelwidth: vte::image::Image width in pixels
+ * @pixelheight: vte::image::Image height in pixels
+ * @left: Left position of image in cell units
+ * @top: Top position of image in cell units
+ * @cell_width: Width of image in cell units
+ * @cell_height: Height of image in cell units
+ *
+ * Append an image to the internal image list.
+ */
+void
+Ring::append_image(vte::Freeable<cairo_surface_t> surface,
+                   int pixelwidth,
+                   int pixelheight,
+                   long left,
+                   long top,
+                   long cell_width,
+                   long cell_height) /* throws */
+{
+        auto const priority = m_next_image_priority;
+        auto [it, success] = m_image_map.try_emplace
+                (priority, // key
+                 std::make_unique<vte::image::Image>(std::move(surface),
+                                                     priority,
+                                                     pixelwidth,
+                                                     pixelheight,
+                                                     left,
+                                                     top,
+                                                     cell_width,
+                                                     cell_height));
+        if (!success)
+                return;
+
+        ++m_next_image_priority;
+
+        auto const& image = it->second;
+
+        m_image_by_top_map.emplace(std::piecewise_construct,
+                                   std::forward_as_tuple(image->get_top()),
+                                   std::forward_as_tuple(image.get()));
+
+        m_image_fast_memory_used += image->resource_size ();
+
+        image_gc_region();
+        image_gc();
+}
+
+#endif /* WITH_SIXEL */
